@@ -15,14 +15,16 @@
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """imports checkers for Python code"""
 
+import collections
+import os
 import sys
-from collections import defaultdict, Counter
 
 import six
 
 import astroid
 from astroid import are_exclusive
-from astroid.modutils import get_module_part, is_standard_module
+from astroid.modutils import (EXT_LIB_DIR, get_module_part, is_standard_module,
+                              file_from_modpath)
 
 from pylint.interfaces import IAstroidChecker
 from pylint.utils import EmptyReport, get_global_option
@@ -32,6 +34,18 @@ from pylint.graph import get_cycles, DotBackend
 from pylint.reporters.ureports.nodes import VerbatimText, Paragraph
 
 
+def _qualified_names(modname):
+    """Split the names of the given module into subparts
+
+    For example,
+        _qualified_names('pylint.checkers.ImportsChecker')
+    returns
+        ['pylint', 'pylint.checkers', 'pylint.checkers.ImportsChecker']
+    """
+    names = modname.split('.')
+    return ['.'.join(names[0:i+1]) for i in range(len(names))]
+
+
 def _get_import_name(importnode, modname):
     """Get a prepared module name from the given import node
 
@@ -49,7 +63,7 @@ def _get_import_name(importnode, modname):
     return modname
 
 
-def get_first_import(node, context, name, base, level):
+def _get_first_import(node, context, name, base, level):
     """return the node where [base.]<name> is imported or None if not found
     """
     fullname = '%s.%s' % (base, name) if base else name
@@ -76,7 +90,7 @@ def get_first_import(node, context, name, base, level):
 
 # utilities to represents import dependencies as tree and dot graph ###########
 
-def make_tree_defs(mod_files_list):
+def _make_tree_defs(mod_files_list):
     """get a list of 2-uple (module, list_of_files_which_import_this_module),
     it will return a dictionary to represent this as a tree
     """
@@ -88,7 +102,8 @@ def make_tree_defs(mod_files_list):
         node[1] += files
     return tree_defs
 
-def repr_tree_defs(data, indent_str=None):
+
+def _repr_tree_defs(data, indent_str=None):
     """return a string which represents imports as a tree"""
     lines = []
     nodes = data.items()
@@ -107,11 +122,11 @@ def repr_tree_defs(data, indent_str=None):
             else:
                 sub_indent_str = '%s| ' % indent_str
         if sub:
-            lines.append(repr_tree_defs(sub, sub_indent_str))
+            lines.append(_repr_tree_defs(sub, sub_indent_str))
     return '\n'.join(lines)
 
 
-def dependencies_graph(filename, dep_info):
+def _dependencies_graph(filename, dep_info):
     """write dependencies as a dot (graphviz) file
     """
     done = {}
@@ -130,11 +145,11 @@ def dependencies_graph(filename, dep_info):
     printer.generate(filename)
 
 
-def make_graph(filename, dep_info, sect, gtype):
+def _make_graph(filename, dep_info, sect, gtype):
     """generate a dependencies graph and add some information about it in the
     report's section
     """
-    dependencies_graph(filename, dep_info)
+    _dependencies_graph(filename, dep_info)
     sect.append(Paragraph('%simports graph has been written to %s'
                           % (gtype, filename)))
 
@@ -178,6 +193,17 @@ MSGS = {
               'multiple-imports',
               'Used when import statement importing multiple modules is '
               'detected.'),
+    'C0411': ('%s comes before %s',
+              'wrong-import-order',
+              'Used when PEP8 import order is not respected (standard imports '
+              'first, then third-party libraries, then local imports)'),
+    'C0412': ('Imports from package %s are not grouped',
+              'ungrouped-imports',
+              'Used when imports are not grouped by packages'),
+    'C0413': ('Import "%s" should be placed at the top of the '
+              'module',
+              'wrong-import-position',
+              'Used when code and imports are mixed'),
     }
 
 class ImportsChecker(BaseChecker):
@@ -194,7 +220,7 @@ class ImportsChecker(BaseChecker):
     msgs = MSGS
     priority = -2
 
-    if sys.version_info < (3,):
+    if six.PY2:
         deprecated_modules = ('regsub', 'TERMIOS', 'Bastion', 'rexec')
     else:
         deprecated_modules = ('optparse', )
@@ -227,16 +253,19 @@ given file (report RP0402 must not be disabled)'}
 given file (report RP0402 must not be disabled)'}
                ),
               )
+    ext_lib_dir = os.path.normcase(os.path.abspath(EXT_LIB_DIR))
 
     def __init__(self, linter=None):
         BaseChecker.__init__(self, linter)
         self.stats = None
         self.import_graph = None
+        self._imports_stack = []
+        self._first_non_import_node = None
         self.__int_dep_info = self.__ext_dep_info = None
         self.reports = (('RP0401', 'External dependencies',
-                         self.report_external_dependencies),
+                         self._report_external_dependencies),
                         ('RP0402', 'Modules dependencies graph',
-                         self.report_dependencies_graph),
+                         self._report_dependencies_graph),
                        )
 
     def open(self):
@@ -244,7 +273,7 @@ given file (report RP0402 must not be disabled)'}
         self.linter.add_stats(dependencies={})
         self.linter.add_stats(cycles=[])
         self.stats = self.linter.stats
-        self.import_graph = defaultdict(set)
+        self.import_graph = collections.defaultdict(set)
         self._ignored_modules = get_global_option(
             self, 'ignored-modules', default=[])
 
@@ -256,26 +285,97 @@ given file (report RP0402 must not be disabled)'}
             for cycle in get_cycles(self.import_graph, vertices=vertices):
                 self.add_message('cyclic-import', args=' -> '.join(cycle))
 
+    @check_messages('wrong-import-position', 'multiple-imports',
+                    'relative-import', 'reimported')
     def visit_import(self, node):
         """triggered when an import statement is seen"""
+        self._check_reimport(node)
+
         modnode = node.root()
         names = [name for name, _ in node.names]
         if len(names) >= 2:
             self.add_message('multiple-imports', args=', '.join(names), node=node)
+
         for name in names:
             self._check_deprecated_module(node, name)
             importedmodnode = self.get_imported_module(node, name)
+            if isinstance(node.scope(), astroid.Module):
+                self._check_position(node)
+                self._record_import(node, importedmodnode)
+
             if importedmodnode is None:
                 continue
+
             self._check_relative_import(modnode, node, importedmodnode, name)
             self._add_imported_module(node, importedmodnode.name)
-            self._check_reimport(node, name)
 
-    # TODO This appears to be the list of all messages of the checker...
-    # @check_messages('W0410', 'W0401', 'W0403', 'W0402', 'W0404', 'W0406', 'E0401')
     @check_messages(*(MSGS.keys()))
     def visit_importfrom(self, node):
         """triggered when a from statement is seen"""
+        basename = node.modname
+        self._check_misplaced_future(node)
+        self._check_deprecated_module(node, basename)
+        self._check_wildcard_imports(node)
+        self._check_same_line_imports(node)
+        self._check_reimport(node, basename=basename, level=node.level)
+
+        modnode = node.root()
+        importedmodnode = self.get_imported_module(node, basename)
+        if isinstance(node.scope(), astroid.Module):
+            self._check_position(node)
+            self._record_import(node, importedmodnode)
+        if importedmodnode is None:
+            return
+        self._check_relative_import(modnode, node, importedmodnode, basename)
+
+        for name, _ in node.names:
+            if name != '*':
+                self._add_imported_module(node, '%s.%s' % (importedmodnode.name, name))
+
+    @check_messages('wrong-import-order', 'ungrouped-imports',
+                    'wrong-import-position')
+    def leave_module(self, node):
+        # Check imports are grouped by category (standard, 3rd party, local)
+        std_imports, ext_imports, loc_imports = self._check_imports_order(node)
+
+        # Check imports are grouped by package within a given category
+        met = set()
+        current_package = None
+        for import_node, import_name in std_imports + ext_imports + loc_imports:
+            package, _, _ = import_name.partition('.')
+            if current_package and current_package != package and package in met:
+                self.add_message('ungrouped-imports', node=import_node,
+                                 args=package)
+            current_package = package
+            met.add(package)
+
+        self._imports_stack = []
+        self._first_non_import_node = None
+
+    def visit_if(self, node):
+        # if the node does not contain an import instruction, and if it is the
+        # first node of the module, keep a track of it (all the import positions
+        # of the module will be compared to the position of this first
+        # instruction)
+        if self._first_non_import_node:
+            return
+        if not isinstance(node.parent, astroid.Module):
+            return
+        if any(node.nodes_of_class((astroid.Import, astroid.ImportFrom))):
+            return
+        self._first_non_import_node = node
+
+    visit_tryfinally = visit_tryexcept = visit_assignattr = visit_assign \
+            = visit_ifexp = visit_comprehension = visit_if
+
+    def visit_functiondef(self, node):
+        # if it is the first non import instruction of the module, record it
+        if not self._first_non_import_node:
+            self._first_non_import_node = node
+
+    visit_classdef = visit_for = visit_while = visit_functiondef
+
+    def _check_misplaced_future(self, node):
         basename = node.modname
         if basename == '__future__':
             # check if this is the first non-docstring statement in the module
@@ -286,29 +386,78 @@ given file (report RP0402 must not be disabled)'}
                         and prev.modname == '__future__'):
                     self.add_message('misplaced-future', node=node)
             return
-        self._check_deprecated_module(node, basename)
-        for name, _ in node.names:
-            if name == '*':
-                self.add_message('wildcard-import', args=basename, node=node)
-        modnode = node.root()
-        importedmodnode = self.get_imported_module(node, basename)
-        if importedmodnode is None:
-            return
-        self._check_relative_import(modnode, node, importedmodnode, basename)
-
-        for name, _ in node.names:
-            if name != '*':
-                self._add_imported_module(node, '%s.%s' % (importedmodnode.name, name))
-                self._check_reimport(node, name, basename, node.level)
 
+    def _check_same_line_imports(self, node):
         # Detect duplicate imports on the same line.
         names = (name for name, _ in node.names)
-        counter = Counter(names)
+        counter = collections.Counter(names)
         for name, count in counter.items():
             if count > 1:
                 self.add_message('reimported', node=node,
                                  args=(name, node.fromlineno))
 
+    def _check_position(self, node):
+        """Check `node` import or importfrom node position is correct
+
+        Send a message  if `node` comes before another instruction
+        """
+        # if a first non-import instruction has already been encountered,
+        # it means the import comes after it and therefore is not well placed
+        if self._first_non_import_node:
+            self.add_message('wrong-import-position', node=node,
+                             args=node.as_string())
+
+    def _record_import(self, node, importedmodnode):
+        """Record the package `node` imports from"""
+        importedname = importedmodnode.name if importedmodnode else None
+        if not importedname:
+            importedname = node.names[0][0].split('.')[0]
+        self._imports_stack.append((node, importedname))
+
+    def _check_imports_order(self, node):
+        """Checks imports of module `node` are grouped by category
+
+        Imports must follow this order: standard, 3rd party, local
+        """
+        extern_imports = []
+        local_imports = []
+        std_imports = []
+        stdlib_paths = [sys.prefix, self.ext_lib_dir]
+        real_prefix = getattr(sys, 'real_prefix', None)
+        if real_prefix is not None:
+            stdlib_paths.append(real_prefix)
+
+        for node, modname in self._imports_stack:
+            package = modname.split('.')[0]
+            if is_standard_module(modname):
+                std_imports.append((node, package))
+                wrong_import = extern_imports or local_imports
+                if not wrong_import:
+                    continue
+                self.add_message('wrong-import-order', node=node,
+                                 args=('standard import "%s"' % node.as_string(),
+                                       '"%s"' % wrong_import[0][0].as_string()))
+            else:
+                try:
+                    filename = file_from_modpath([package])
+                except ImportError:
+                    extern_imports.append((node, package))
+                    continue
+                if not filename:
+                    extern_imports.append((node, package))
+                    continue
+                filename = os.path.normcase(os.path.abspath(filename))
+                if not any(filename.startswith(path) for path in stdlib_paths):
+                    local_imports.append((node, package))
+                    continue
+                extern_imports.append((node, package))
+                if not local_imports:
+                    continue
+                self.add_message('wrong-import-order', node=node,
+                                 args=('external import "%s"' % node.as_string(),
+                                       '"%s"' % local_imports[0][0].as_string()))
+        return std_imports, extern_imports, local_imports
+
     def get_imported_module(self, importnode, modname):
         try:
             return importnode.do_import_module(modname)
@@ -319,25 +468,13 @@ given file (report RP0402 must not be disabled)'}
             else:
                 args = repr(dotted_modname)
 
-            for submodule in self._qualified_names(modname):
+            for submodule in _qualified_names(modname):
                 if submodule in self._ignored_modules:
                     return None
 
             if not node_ignores_exception(importnode, ImportError):
                 self.add_message("import-error", args=args, node=importnode)
 
-    @staticmethod
-    def _qualified_names(modname):
-        """Split the names of the given module into subparts
-
-        For example,
-            _qualified_names('pylint.checkers.ImportsChecker')
-        returns
-            ['pylint', 'pylint.checkers', 'pylint.checkers.ImportsChecker']
-        """
-        names = modname.split('.')
-        return ['.'.join(names[0:i+1]) for i in range(len(names))]
-
     def _check_relative_import(self, modnode, importnode, importedmodnode,
                                importedasname):
         """check relative import. node is either an Import or From node, modname
@@ -359,20 +496,32 @@ given file (report RP0402 must not be disabled)'}
 
     def _add_imported_module(self, node, importedmodname):
         """notify an imported module, used to analyze dependencies"""
+        module_file = node.root().file
+        context_name = node.root().name
+        base = os.path.splitext(os.path.basename(module_file))[0]
+
+        # Determine if we have a `from .something import` in a package's
+        # __init__. This means the module will never be able to import
+        # itself using this condition (the level will be bigger or
+        # if the same module is named as the package, it will be different
+        # anyway).
+        if isinstance(node, astroid.ImportFrom):
+            if node.level and node.level > 0 and base == '__init__':
+                return
+
         try:
             importedmodname = get_module_part(importedmodname,
-                                              node.root().file)
+                                              module_file)
         except ImportError:
             pass
-        context_name = node.root().name
+
         if context_name == importedmodname:
-            # module importing itself !
             self.add_message('import-self', node=node)
         elif not is_standard_module(importedmodname):
             # handle dependencies
             importedmodnames = self.stats['dependencies'].setdefault(
                 importedmodname, set())
-            if not context_name in importedmodnames:
+            if context_name not in importedmodnames:
                 importedmodnames.add(context_name)
             # update import graph
             mgraph = self.import_graph[context_name]
@@ -385,7 +534,7 @@ given file (report RP0402 must not be disabled)'}
             if mod_path == mod_name or mod_path.startswith(mod_name + '.'):
                 self.add_message('deprecated-module', node=node, args=mod_path)
 
-    def _check_reimport(self, node, name, basename=None, level=None):
+    def _check_reimport(self, node, basename=None, level=None):
         """check if the import is necessary (i.e. not already done)"""
         if not self.linter.is_message_enabled('reimported'):
             return
@@ -395,22 +544,23 @@ given file (report RP0402 must not be disabled)'}
         contexts = [(frame, level)]
         if root is not frame:
             contexts.append((root, None))
+
         for context, level in contexts:
-            first = get_first_import(node, context, name, basename, level)
-            if first is not None:
-                self.add_message('reimported', node=node,
-                                 args=(name, first.fromlineno))
+            for name, _ in node.names:
+                first = _get_first_import(node, context, name, basename, level)
+                if first is not None:
+                    self.add_message('reimported', node=node,
+                                     args=(name, first.fromlineno))
 
-
-    def report_external_dependencies(self, sect, _, dummy):
+    def _report_external_dependencies(self, sect, _, dummy):
         """return a verbatim layout for displaying dependencies"""
-        dep_info = make_tree_defs(six.iteritems(self._external_dependencies_info()))
+        dep_info = _make_tree_defs(six.iteritems(self._external_dependencies_info()))
         if not dep_info:
             raise EmptyReport()
-        tree_str = repr_tree_defs(dep_info)
+        tree_str = _repr_tree_defs(dep_info)
         sect.append(VerbatimText(tree_str))
 
-    def report_dependencies_graph(self, sect, _, dummy):
+    def _report_dependencies_graph(self, sect, _, dummy):
         """write dependencies as a dot (graphviz) file"""
         dep_info = self.stats['dependencies']
         if not dep_info or not (self.config.import_graph
@@ -419,15 +569,15 @@ given file (report RP0402 must not be disabled)'}
             raise EmptyReport()
         filename = self.config.import_graph
         if filename:
-            make_graph(filename, dep_info, sect, '')
+            _make_graph(filename, dep_info, sect, '')
         filename = self.config.ext_import_graph
         if filename:
-            make_graph(filename, self._external_dependencies_info(),
-                       sect, 'external ')
+            _make_graph(filename, self._external_dependencies_info(),
+                        sect, 'external ')
         filename = self.config.int_import_graph
         if filename:
-            make_graph(filename, self._internal_dependencies_info(),
-                       sect, 'internal ')
+            _make_graph(filename, self._internal_dependencies_info(),
+                        sect, 'internal ')
 
     def _external_dependencies_info(self):
         """return cached external dependencies information or build and
@@ -453,6 +603,11 @@ given file (report RP0402 must not be disabled)'}
                     result[importee] = importers
         return self.__int_dep_info
 
+    def _check_wildcard_imports(self, node):
+        for name, _ in node.names:
+            if name == '*':
+                self.add_message('wildcard-import', args=node.modname, node=node)
+
 
 def register(linter):
     """required method to auto register this checker """
